-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AloftCombatText")
local SML = AceLibrary("SharedMedia-1.0")
local SML_font = SML:List("font")

-----------------------------------------------------------------------------

L:RegisterTranslations("enUS", function() return {
	["Combat Text"] = true,
	["Combat text options"] = true,
	["Enable Group"] = true,
	["Show combat text on group member nameplates"] = true,
	["Enable Target"] = true,
	["Show combat text on target nameplates"] = true,

	["Typeface"] = true,
	["Combat text typeface options"] = true,
	["Font"] = true,
	["Sets the font for combat text"] = true,
	["Font Size"] = true,
	["Sets the font height of the combat text"] = true,
	["Font Shadow"] = true,
	["Show font shadow on combat text"] = true,
	["Outline"] = true,
	["Sets the outline for combat text"] = true,

	["Position"] = true,
	["Adjust combat text position"] = true,
	["Anchor"] = true, 
	["Sets the anchor for the combat text"] = true,
	["Anchor To"] = true, 
	["Sets the relative point on the health bar to anchor the combat text"] = true,
	["X Offset"] = true,
	["X offset of the combat text"] = true,
	["Y Offset"] = true,
	["Y offset of the combat text"] = true,
	["Alpha"] = true,
	["Sets the alpha of the combat text"] = true,

	["None"] = true,
	["Normal"] = true,
	["Thick"] = true,

	["Colors"] = true,
	["Customize colors"] = true,
	["Physical"] = true,
	["Sets the color for physical damage"] = true,
	["Holy"] = true,
	["Sets the color for holy damage"] = true,
	["Fire"] = true,
	["Sets the color for fire damage"] = true,
	["Nature"] = true,
	["Sets the color for nature damage"] = true,
	["Frost"] = true,
	["Sets the color for frost damage"] = true,
	["Shadow"] = true,
	["Sets the color for shadow damage"] = true,
	["Arcane"] = true,
	["Sets the color for arcane damage"] = true,
	["Heal"] = true,
	["Sets the color for healing"] = true,
	["Text"] = true,
	["Sets the color for text feedback"] = true,
	["Reset"] = true,
	["Resets all of the colors to default"] = true,
	["Set All"] = true,
	["Sets all of the colors"] = true,
} end)

-----------------------------------------------------------------------------

local AloftCombatText = Aloft:NewModule("CombatText", "AceEvent-2.0")
AloftCombatText.db = Aloft:AcquireDBNamespace("combatText")
Aloft:RegisterDefaults("combatText", "profile", {
	enable			= true, -- not used for any module control logic, just required for PlaceFrame and PrepareText to work
	enableGroup		= false,
	enableTarget	= false,
	alpha			= 0.75,
	point			= "CENTER",
	relativeToPoint	= "CENTER",
	offsetX			= 0,
	offsetY			= 0,
	font			= "Friz Quadrata TT",
	fontSize		= 16,
	shadow			= false,
	outline			= "OUTLINE",
	colors =
	{
		[0] = { 1, 1, 1 },			-- Physical
		[1] = { 1, 1, 0.5 },  		-- Holy
		[2] = { 1, 0.3, 0 },		-- Fire
		[3] = { 0.5, 1, 0.2 },		-- Nature
		[4] = { 0.4, 0.6, 0.9 },	-- Frost,
		[5] = { 0.7, 0.4, 1 },	-- Shadow
		[6] = { 0.8, 0.8, 1 },		-- Arcane
		energize = { 0.41, 0.8, 0.94 },
		text = { 1, 1, 1 },
		heal = { 0, 1, 0 },
	},
})

Aloft.Options.args.combatText =
{
	type = 'group',
	name = L["Combat Text"],
	desc = L["Combat text options"],
	disabled = function() return not Aloft:IsActive() end,
	args =
	{
		enableGroup =
		{
			type = 'toggle',
			name = L["Enable Group"],
			desc = L["Show combat text on group member nameplates"],
			get = function() return AloftCombatText.db.profile.enableGroup end,
			set = function(v) AloftCombatText.db.profile.enableGroup = v AloftCombatText:UpdateAll() end,
			order = 1
		},
		enableTarget =
		{
			type = 'toggle',
			name = L["Enable Target"],
			desc = L["Show combat text on target nameplates"],
			get = function() return AloftCombatText.db.profile.enableTarget end,
			set = function(v) AloftCombatText.db.profile.enableTarget = v AloftCombatText:UpdateAll() end,
			order = 1
		},
		typeface =
		{
			type = "group",
			name = L["Typeface"],
			desc = L["Combat text typeface options"],
			disabled = function() return not AloftCombatText.db.profile.enableGroup and not AloftCombatText.db.profile.enableTarget end,
			args =
			{
				font =
				{
					type = "text", 
					name = L["Font"], 
					desc = L["Sets the font for combat text"],
					get = function() return AloftCombatText.db.profile.font end,
					set = function(v) AloftCombatText.db.profile.font = v AloftCombatText:UpdateAll() end,
					validate = SML_font,
				},
				fontSize =
				{
					type = 'range',
					name = L["Font Size"],
					desc = L["Sets the font height of the combat text"],
					max = 40,
					min = 5,
					step = 1,
					get = function() return AloftCombatText.db.profile.fontSize end,
					set = function(value) AloftCombatText.db.profile.fontSize = value AloftCombatText:UpdateAll() end
				},
				shadow =
				{
					type = 'toggle',
					name = L["Font Shadow"],
					desc = L["Show font shadow on combat text"],
					get = function() return AloftCombatText.db.profile.shadow end,
					set = function(v) AloftCombatText.db.profile.shadow = v AloftCombatText:UpdateAll() end
				},
				outline =
				{
					type = 'text',
					name = L["Outline"],
					desc = L["Sets the outline for combat text"],
					get = function() return AloftCombatText.db.profile.outline end,
					set = function(value) AloftCombatText.db.profile.outline = value AloftCombatText:UpdateAll() end,
					validate = { [""] = L["None"], ["OUTLINE"] = L["Normal"], ["THICKOUTLINE"] = L["Thick"] },
				},
			},
		},
		position =
		{
			type = 'group',
			name = L["Position"],
			desc = L["Adjust combat text position"],
			disabled = function() return not AloftCombatText.db.profile.enableGroup and not AloftCombatText.db.profile.enableTarget end,
			args =
			{
				anchor = {
					type = "text", 
					name = L["Anchor"], 
					desc = L["Sets the anchor for the combat text"],
					get = function() return AloftCombatText.db.profile.point end,
					set = function(v) AloftCombatText.db.profile.point = v AloftCombatText:UpdateAll() end,
					validate = {"TOPLEFT", "TOP", "TOPRIGHT", "LEFT", "CENTER", "RIGHT", "BOTTOMLEFT", "BOTTOM", "BOTTOMRIGHT"},
				},
				anchorto = {
					type = "text", 
					name = L["Anchor To"], 
					desc = L["Sets the relative point on the health bar to anchor the combat text"],
					get = function() return AloftCombatText.db.profile.relativeToPoint end,
					set = function(v) AloftCombatText.db.profile.relativeToPoint = v AloftCombatText:UpdateAll() end,
					validate = {"TOPLEFT", "TOP", "TOPRIGHT", "LEFT", "CENTER", "RIGHT", "BOTTOMLEFT", "BOTTOM", "BOTTOMRIGHT"},
				},
				offsetX =
				{
					type = 'range',
					name = L["X Offset"],
					desc = L["X offset of the combat text"],
					min = -32,
					max = 32,
					step = 1,
					get = function() return AloftCombatText.db.profile.offsetX end,
					set = function(v) AloftCombatText.db.profile.offsetX = v AloftCombatText:UpdateAll() end
				},
				offsetY =
				{
					type = 'range',
					name = L["Y Offset"],
					desc = L["Y offset of the combat text"],
					min = -32,
					max = 32,
					step = 1,
					get = function() return AloftCombatText.db.profile.offsetY end,
					set = function(v) AloftCombatText.db.profile.offsetY = v AloftCombatText:UpdateAll() end
				},
			},
		},
		alpha =
		{
			type = 'range',
			name = L["Alpha"],
			desc = L["Sets the alpha of the combat text"],
			min = 0,
			max = 1.0,
			step = 0.05,
			disabled = function() return not AloftCombatText.db.profile.enableGroup and not AloftCombatText.db.profile.enableTarget end,
			get = function() return AloftCombatText.db.profile.alpha end,
			set = function(v) AloftCombatText.db.profile.alpha = v AloftCombatText:UpdateAll() end
		},
		colors =
		{
			type = 'group',
			name = L["Colors"],
			desc = L["Customize colors"],
			disabled = function() return not AloftCombatText.db.profile.enableGroup and not AloftCombatText.db.profile.enableTarget end,
			args =
			{
				physical =
				{
					type = 'color',
					name = L["Physical"],
					desc = L["Sets the color for physical damage"],
					order = 1,
					get = function() return unpack(AloftCombatText.db.profile.colors[0]) end,
					set = function(r, g, b) AloftCombatText.db.profile.colors[0] = { r, g, b } end,
				},
				holy =
				{
					type = 'color',
					name = L["Holy"],
					desc = L["Sets the color for holy damage"],
					order = 2,
					get = function() return unpack(AloftCombatText.db.profile.colors[1]) end,
					set = function(r, g, b) AloftCombatText.db.profile.colors[1] = { r, g, b } end,
				},
				fire =
				{
					type = 'color',
					name = L["Fire"],
					desc = L["Sets the color for fire damage"],
					order = 2,
					get = function() return unpack(AloftCombatText.db.profile.colors[2]) end,
					set = function(r, g, b) AloftCombatText.db.profile.colors[2] = { r, g, b } end,
				},
				nature =
				{
					type = 'color',
					name = L["Nature"],
					desc = L["Sets the color for nature damage"],
					order = 2,
					get = function() return unpack(AloftCombatText.db.profile.colors[3]) end,
					set = function(r, g, b) AloftCombatText.db.profile.colors[3] = { r, g, b } end,
				},
				frost =
				{
					type = 'color',
					name = L["Frost"],
					desc = L["Sets the color for frost damage"],
					order = 2,
					get = function() return unpack(AloftCombatText.db.profile.colors[4]) end,
					set = function(r, g, b) AloftCombatText.db.profile.colors[4] = { r, g, b } end,
				},
				shadow =
				{
					type = 'color',
					name = L["Shadow"],
					desc = L["Sets the color for shadow damage"],
					order = 2,
					get = function() return unpack(AloftCombatText.db.profile.colors[5]) end,
					set = function(r, g, b) AloftCombatText.db.profile.colors[5] = { r, g, b } end,
				},
				arcane =
				{
					type = 'color',
					name = L["Arcane"],
					desc = L["Sets the color for arcane damage"],
					order = 2,
					get = function() return unpack(AloftCombatText.db.profile.colors[6]) end,
					set = function(r, g, b) AloftCombatText.db.profile.colors[6] = { r, g, b } end,
				},
				heal =
				{
					type = 'color',
					name = L["Heal"],
					desc = L["Sets the color for healing"],
					order = 3,
					get = function() return unpack(AloftCombatText.db.profile.colors.heal) end,
					set = function(r, g, b) AloftCombatText.db.profile.colors.heal = { r, g, b } end,
				},
				text =
				{
					type = 'color',
					name = L["Text"],
					desc = L["Sets the color for text feedback"],
					order = 3,
					get = function() return unpack(AloftCombatText.db.profile.colors.text) end,
					set = function(r, g, b) AloftCombatText.db.profile.colors.text = { r, g, b } end,
				},
				reset =
				{
					type = 'execute',
					name = L["Reset"],
					desc = L["Resets all of the colors to default"],
					order = -1,
					func = function()
								AloftCombatText.db.profile.colors[0] = { 1, 1, 1 }			-- Physical
								AloftCombatText.db.profile.colors[1] = { 1, 1, 0.5 }  		-- Holy
								AloftCombatText.db.profile.colors[2] = { 1, 0.3, 0 }		-- Fire
								AloftCombatText.db.profile.colors[3] = { 0.5, 1, 0.2 }		-- Nature
								AloftCombatText.db.profile.colors[4] = { 0.4, 0.6, 0.9 }	-- Frost,
								AloftCombatText.db.profile.colors[5] = { 0.7, 0.4, 1 }		-- Shadow
								AloftCombatText.db.profile.colors[6] = { 0.8, 0.8, 1 }		-- Arcane
								AloftCombatText.db.profile.colors.energize = { 0.41, 0.8, 0.94 }
								AloftCombatText.db.profile.colors.text = { 1, 1, 1 }
								AloftCombatText.db.profile.colors.heal = { 0, 1, 0 }
							end
				},
				setAll =
				{
					type = 'color',
					name = L["Set All"],
					desc = L["Sets all of the colors"],
					get = function() return unpack(AloftCombatText.db.profile.colors[0]) end,
					set = function(r, g, b)
							AloftCombatText.db.profile.colors[0] = { r, g, b } 
							AloftCombatText.db.profile.colors[1] = { r, g, b } 
							AloftCombatText.db.profile.colors[2] = { r, g, b } 
							AloftCombatText.db.profile.colors[3] = { r, g, b } 
							AloftCombatText.db.profile.colors[4] = { r, g, b } 
							AloftCombatText.db.profile.colors[5] = { r, g, b } 
							AloftCombatText.db.profile.colors[6] = { r, g, b } 
							AloftCombatText.db.profile.colors.energize = { r, g, b }
							AloftCombatText.db.profile.colors.text = { r, g, b }
							AloftCombatText.db.profile.colors.heal = { r, g, b }
						  end,
					order = -1,
				},
			},
		},
	},
}

-----------------------------------------------------------------------------

local textRegionPool = { }
local updateList = { }
local colors

-----------------------------------------------------------------------------

function AloftCombatText:UpdateAll()
	for aloftData in Aloft:IterateNameplates() do
		if self.db.profile.enableTarget or self.db.profile.enableGroup then
			if aloftData.combatTextRegion then
				self:PlaceFrame(aloftData.combatTextRegion, aloftData.healthBar, self.db.profile)
				self:PrepareText(aloftData.combatTextRegion, self.db.profile)
			end
		else
			self:OnNameplateHide(aloftData)
		end
	end
	for textRegion in pairs(textRegionPool) do
		self:PrepareText(textRegion, self.db.profile)
	end
	self:RegisterEvents()
end

-----------------------------------------------------------------------------

function AloftCombatText:RegisterEvents()
	local profile = self.db.profile
	self:UnregisterAllEvents()

	self:RegisterEvent("Aloft:SetAll", function(type, value)
		if self.db.profile[type] then
			self.db.profile[type] = value
			self:UpdateAll()
		end
	end)
	
	if profile.enableGroup or profile.enableTarget then
		self:RegisterEvent("UNIT_COMBAT", "OnUnitCombat")
		self:RegisterEvent("Aloft:OnNameplateHide", "OnNameplateHide")
	end
end

function AloftCombatText:RequiresData()
	if self.db.profile.enableGroup then
		return "unitid"
	end
end

function AloftCombatText:OnEnable()
	colors = self.db.profile.colors
	self:UpdateAll()
end

function AloftCombatText:OnDisable()
	self:UnregisterAllEvents()
	
	for aloftData in Aloft:IterateVisibleNameplates() do
		self:OnNameplateHide(aloftData)
	end
end

-----------------------------------------------------------------------------

-- These return text, scale, r, g, b
local combatHandlers =
{
	["BLOCK"]		= function() return CombatFeedbackText["BLOCK"], 0.75, colors.text end,
	["ENERGIZE"]	= function(flags, amount) return amount, flags == "CRITICAL" and 1.3 or 1.0, colors.energize end,
	["HEAL"]		= function(flags, amount) return "+"..amount, flags == "CRITICAL" and 1.3 or 1.0, colors.heal end,
	["IMMUNE"]		= function() return CombatFeedbackText["IMMUNE"], 0.5, colors.text end,
	["WOUND"]		= function(flags, amount, type) 
						if amount ~= 0 then
							local scale = 1
							if flags == "CRITICAL" or flags == "CRUSHING" then
								scale = 1.5
							elseif flags == "GLANCING" then
								scale = 0.75
							end
							return "-"..amount, scale, colors[type]
						elseif flags == "ABSORB" or flags == "BLOCK" or flags == "RESIST" then
							return CombatFeedbackText[flags], 0.75, colors.text
						end
						return CombatFeedbackText[flags], 1.0, colors.text
					  end,
}

function AloftCombatText:OnUnitCombat(unit, event, ...)	-- last 3 arguments are flags, amount, type
	if unit == "player" then return end
	
	local aloftData	
	local profile = self.db.profile
	
	if unit == "target" then
		if not profile.enableTarget then return end
		aloftData = Aloft:GetTargetNameplate()
	else
		if not profile.enableGroup then return end
		aloftData = Aloft:GetNameplate(UnitName(unit))	
	end
	if not aloftData then return end

	local text, scale, color
	if combatHandlers[event] then
		text, scale, color = combatHandlers[event](...)
	else
		text, scale, color = CombatFeedbackText[event], 1.0, colors.text
	end

    local combatTextRegion = aloftData.combatTextRegion    
    if not combatTextRegion then
        combatTextRegion = next(textRegionPool)
        if combatTextRegion then
            textRegionPool[combatTextRegion] = nil
            combatTextRegion:SetParent(aloftData.nameplateFrame)
        else
            -- We need a new text region.
            combatTextRegion = aloftData:CreateFontString()
	        self:PrepareText(combatTextRegion, profile)
        end
        self:PlaceFrame(combatTextRegion, aloftData.healthBar, profile)
        aloftData.combatTextRegion = combatTextRegion
    end

	-- The line below crashes the Windows WoW client sporadically (not the Mac one). SetTextHeight is used instead.
	--combatTextRegion:SetFont(SML:Fetch("font", profile.font), profile.fontSize * scale, profile.outline)
	combatTextRegion:SetTextHeight(profile.fontSize * scale)
    combatTextRegion:SetText(text)
    combatTextRegion:SetTextColor(unpack(color))
    combatTextRegion:SetAlpha(0)
    combatTextRegion:Show()

    updateList[aloftData] = GetTime()

	self:ScheduleRepeatingEvent("AloftCombatText:Update", self.Update, 0.05, self)
end

function AloftCombatText:Update()
	local now = GetTime()
	
	for aloftData, startTime in pairs(updateList) do
		local elapsedTime = now - startTime
		local combatTextRegion = aloftData.combatTextRegion

		if elapsedTime < COMBATFEEDBACK_FADEINTIME then
			combatTextRegion:SetAlpha(self.db.profile.alpha*(elapsedTime / COMBATFEEDBACK_FADEINTIME))
		elseif elapsedTime < (COMBATFEEDBACK_FADEINTIME + COMBATFEEDBACK_HOLDTIME) then
			combatTextRegion:SetAlpha(self.db.profile.alpha)
		elseif elapsedTime < (COMBATFEEDBACK_FADEINTIME + COMBATFEEDBACK_HOLDTIME + COMBATFEEDBACK_FADEOUTTIME) then
			combatTextRegion:SetAlpha(self.db.profile.alpha * (1 - (elapsedTime - COMBATFEEDBACK_HOLDTIME - COMBATFEEDBACK_FADEINTIME) / COMBATFEEDBACK_FADEOUTTIME))
		else
			combatTextRegion:Hide()
			textRegionPool[combatTextRegion] = true
			aloftData.combatTextRegion = nil
			updateList[aloftData] = nil
		end		
	end

    if not next(updateList) then
		self:CancelScheduledEvent("AloftCombatText:Update")
    end
end

function AloftCombatText:OnNameplateHide(aloftData)
    local combatTextRegion = aloftData.combatTextRegion
    if combatTextRegion then
        combatTextRegion:Hide()
		textRegionPool[combatTextRegion] = true
		aloftData.combatTextRegion = nil
        updateList[aloftData] = nil
        
        if not next(updateList) then
			self:CancelScheduledEvent("AloftCombatText:Update")
        end
    end
end

-----------------------------------------------------------------------------
